@@ -31,6 +31,7 @@ from gnocchi import aggregates
 from gnocchi import archive_policy
 from gnocchi import indexer
 from gnocchi import json
+from gnocchi import resource_type
 from gnocchi import storage
 from gnocchi import utils
 
@@ -81,7 +82,8 @@ def enforce(rule, target):
     creds = {
         'roles': headers.get("X-Roles", "").split(","),
         'user_id': user_id,
-        'project_id': project_id
+        'project_id': project_id,
+        'domain_id': headers.get("X-Domain-Id", ""),
     }
 
     if not isinstance(target, dict):
@@ -166,8 +168,6 @@ def deserialize_and_validate(schema, required=True):
 
 
 def Timestamp(v):
-    if v is None:
-        return v
     t = utils.to_timestamp(v)
     if t < utils.unix_universal_start:
         raise ValueError("Timestamp must be after Epoch")
@@ -218,6 +218,8 @@ def get_details(params):
 RESOURCE_DEFAULT_PAGINATION = ['revision_start:asc',
                                'started_at:asc']
 
+METRIC_DEFAULT_PAGINATION = ['id:asc']
+
 
 def get_pagination_options(params, default):
     max_limit = pecan.request.conf.api.max_limit
@@ -258,6 +260,33 @@ class ArchivePolicyController(rest.RestController):
             return ap
         abort(404, indexer.NoSuchArchivePolicy(self.archive_policy))
 
+    @pecan.expose('json')
+    def patch(self):
+        ap = pecan.request.indexer.get_archive_policy(self.archive_policy)
+        if not ap:
+            abort(404, indexer.NoSuchArchivePolicy(self.archive_policy))
+        enforce("update archive policy", ap)
+
+        body = deserialize_and_validate(voluptuous.Schema({
+            voluptuous.Required("definition"):
+                voluptuous.All([{
+                    "granularity": Timespan,
+                    "points": PositiveNotNullInt,
+                    "timespan": Timespan}], voluptuous.Length(min=1)),
+            }))
+        # Validate the data
+        try:
+            ap_items = [archive_policy.ArchivePolicyItem(**item) for item in
+                        body['definition']]
+        except ValueError as e:
+            abort(400, e)
+
+        try:
+            return pecan.request.indexer.update_archive_policy(
+                self.archive_policy, ap_items)
+        except indexer.UnsupportedArchivePolicyChange as e:
+            abort(400, e)
+
     @pecan.expose()
     def delete(self):
         # NOTE(jd) I don't think there's any point in fetching and passing the
@@ -568,13 +597,6 @@ class MetricController(rest.RestController):
             abort(404, e)
 
 
-def UUID(value):
-    try:
-        return uuid.UUID(value)
-    except Exception as e:
-        raise ValueError(e)
-
-
 class MetricsController(rest.RestController):
 
     @pecan.expose()
@@ -594,6 +616,8 @@ class MetricsController(rest.RestController):
         "project_id": six.text_type,
         "archive_policy_name": six.text_type,
         "name": six.text_type,
+        voluptuous.Optional("unit"):
+            voluptuous.All(six.text_type, voluptuous.Length(max=31)),
     })
 
     # NOTE(jd) Define this method as it was a voluptuous schema – it's just a
@@ -629,6 +653,7 @@ class MetricsController(rest.RestController):
             "project_id": definition.get('project_id'),
             "archive_policy_name": archive_policy_name,
             "name": name,
+            "unit": definition.get('unit'),
         })
 
         return definition
@@ -642,6 +667,7 @@ class MetricsController(rest.RestController):
                 uuid.uuid4(),
                 user, project,
                 name=body.get('name'),
+                unit=body.get('unit'),
                 archive_policy_name=body['archive_policy_name'])
         except indexer.NoSuchArchivePolicy as e:
             abort(400, e)
@@ -671,11 +697,16 @@ class MetricsController(rest.RestController):
             attr_filter['created_by_user_id'] = user_id
         if project_id is not None:
             attr_filter['created_by_project_id'] = project_id
-        return pecan.request.indexer.list_metrics(**attr_filter)
+        attr_filter.update(get_pagination_options(
+            kwargs, METRIC_DEFAULT_PAGINATION))
+        try:
+            return pecan.request.indexer.list_metrics(**attr_filter)
+        except indexer.IndexerException as e:
+            abort(400, e)
 
 
 _MetricsSchema = voluptuous.Schema({
-    six.text_type: voluptuous.Any(UUID,
+    six.text_type: voluptuous.Any(utils.UUID,
                                   MetricsController.MetricSchema),
 })
 
@@ -759,7 +790,7 @@ class ResourceHistoryController(rest.RestController):
         resource = pecan.request.indexer.get_resource(
             self.resource_type, self.resource_id)
         if not resource:
-            abort(404, "foo")
+            abort(404, indexer.NoSuchResource(self.resource_id))
 
         enforce("get resource", resource)
 
@@ -808,6 +839,68 @@ def etag_set_headers(obj):
     pecan.response.last_modified = obj.lastmodified
 
 
+class ResourceTypeController(rest.RestController):
+    def __init__(self, name):
+        self._name = name
+
+    @pecan.expose('json')
+    def get(self):
+        try:
+            rt = pecan.request.indexer.get_resource_type(self._name)
+        except indexer.NoSuchResourceType as e:
+            abort(404, e)
+        enforce("get resource type", rt)
+        return rt
+
+    @pecan.expose()
+    def delete(self):
+        try:
+            pecan.request.indexer.get_resource_type(self._name)
+        except indexer.NoSuchResourceType as e:
+            abort(404, e)
+        enforce("delete resource type", resource_type)
+        try:
+            pecan.request.indexer.delete_resource_type(self._name)
+        except (indexer.NoSuchResourceType,
+                indexer.ResourceTypeInUse) as e:
+            abort(400, e)
+
+
+class ResourceTypesController(rest.RestController):
+
+    @pecan.expose()
+    def _lookup(self, name, *remainder):
+        return ResourceTypeController(name), remainder
+
+    @pecan.expose('json')
+    def post(self):
+        schema = pecan.request.indexer.get_resource_type_schema()
+        body = deserialize_and_validate(schema)
+        body["state"] = "creating"
+
+        try:
+            rt = schema.resource_type_from_dict(**body)
+        except resource_type.InvalidResourceAttributeName as e:
+            abort(400, e)
+
+        enforce("create resource type", body)
+        try:
+            rt = pecan.request.indexer.create_resource_type(rt)
+        except indexer.ResourceTypeAlreadyExists as e:
+            abort(409, e)
+        set_resp_location_hdr("/resource_type/" + rt.name)
+        pecan.response.status = 201
+        return rt
+
+    @pecan.expose('json')
+    def get_all(self, **kwargs):
+        enforce("list resource type", {})
+        try:
+            return pecan.request.indexer.list_resource_types()
+        except indexer.IndexerException as e:
+            abort(400, e)
+
+
 def ResourceSchema(schema):
     base_schema = {
         voluptuous.Optional('started_at'): Timestamp,
@@ -900,45 +993,9 @@ class ResourceController(rest.RestController):
             abort(404, e)
 
 
-GenericSchema = ResourceSchema({})
-
-InstanceDiskSchema = ResourceSchema({
-    "name": six.text_type,
-    "instance_id": UUID,
-})
-
-InstanceNetworkInterfaceSchema = ResourceSchema({
-    "name": six.text_type,
-    "instance_id": UUID,
-})
-
-InstanceSchema = ResourceSchema({
-    "flavor_id": six.text_type,
-    voluptuous.Optional("image_ref"): six.text_type,
-    "host": six.text_type,
-    "display_name": six.text_type,
-    voluptuous.Optional("server_group"): six.text_type,
-})
-
-VolumeSchema = ResourceSchema({
-    voluptuous.Optional("display_name"): voluptuous.Any(None,
-                                                        six.text_type),
-})
-
-ImageSchema = ResourceSchema({
-    "name": six.text_type,
-    "container_format": six.text_type,
-    "disk_format": six.text_type,
-})
-
-
-# NOTE(sileht): Must be loaded after all ResourceSchema
-RESOURCE_SCHEMA_MANAGER = extension.ExtensionManager(
-    'gnocchi.controller.schemas')
-
-
 def schema_for(resource_type):
-    return RESOURCE_SCHEMA_MANAGER[resource_type].plugin
+    resource_type = pecan.request.indexer.get_resource_type(resource_type)
+    return ResourceSchema(resource_type.schema)
 
 
 def ResourceID(value):
@@ -1016,16 +1073,17 @@ class ResourcesByTypeController(rest.RestController):
     @pecan.expose('json')
     def get_all(self):
         return dict(
-            (ext.name,
-             pecan.request.application_url + '/resource/' + ext.name)
-            for ext in RESOURCE_SCHEMA_MANAGER)
+            (rt.name,
+             pecan.request.application_url + '/resource/' + rt.name)
+            for rt in pecan.request.indexer.list_resource_types())
 
     @pecan.expose()
     def _lookup(self, resource_type, *remainder):
-        if resource_type in RESOURCE_SCHEMA_MANAGER:
-            return ResourcesController(resource_type), remainder
-        else:
-            abort(404, indexer.NoSuchResourceType(resource_type))
+        try:
+            pecan.request.indexer.get_resource_type(resource_type)
+        except indexer.NoSuchResourceType as e:
+            abort(404, e)
+        return ResourcesController(resource_type), remainder
 
 
 def _ResourceSearchSchema(v):
@@ -1101,10 +1159,11 @@ class SearchResourceTypeController(rest.RestController):
 class SearchResourceController(rest.RestController):
     @pecan.expose()
     def _lookup(self, resource_type, *remainder):
-        if resource_type in RESOURCE_SCHEMA_MANAGER:
-            return SearchResourceTypeController(resource_type), remainder
-        else:
-            abort(404, indexer.NoSuchResourceType(resource_type))
+        try:
+            pecan.request.indexer.get_resource_type(resource_type)
+        except indexer.NoSuchResourceType as e:
+            abort(404, e)
+        return SearchResourceTypeController(resource_type), remainder
 
 
 def _MetricSearchSchema(v):
@@ -1247,7 +1306,7 @@ class MetricsMeasuresBatchController(rest.RestController):
     # only the last key will be retain by json python module to
     # build the python dict.
     MeasuresBatchSchema = voluptuous.Schema(
-        {UUID: [MeasureSchema]}
+        {utils.UUID: [MeasureSchema]}
     )
 
     @pecan.expose()
@@ -1337,8 +1396,10 @@ class AggregationController(rest.RestController):
             # NOTE(sileht): we want the raw 404 message here
             # so use directly pecan
             pecan.abort(404)
-        elif resource_type not in RESOURCE_SCHEMA_MANAGER:
-            abort(404, indexer.NoSuchResourceType(resource_type))
+        try:
+            pecan.request.indexer.get_resource_type(resource_type)
+        except indexer.NoSuchResourceType as e:
+            abort(404, e)
         return AggregationResourceController(resource_type,
                                              metric_name), remainder
 
@@ -1357,10 +1418,11 @@ class CapabilityController(rest.RestController):
     def get():
         aggregation_methods = set(
             archive_policy.ArchivePolicy.VALID_AGGREGATION_METHODS)
-        aggregation_methods.update(
-            ext.name for ext in extension.ExtensionManager(
-                namespace='gnocchi.aggregates'))
-        return dict(aggregation_methods=aggregation_methods)
+        return dict(aggregation_methods=aggregation_methods,
+                    dynamic_aggregation_methods=[
+                        ext.name for ext in extension.ExtensionManager(
+                            namespace='gnocchi.aggregates')
+                    ])
 
 
 class StatusController(rest.RestController):
@@ -1368,7 +1430,8 @@ class StatusController(rest.RestController):
     @pecan.expose('json')
     def get(details=True):
         enforce("get status", {})
-        report = pecan.request.storage.measures_report(details)
+        report = pecan.request.storage.measures_report(
+            strutils.bool_from_string(details))
         report_dict = {"storage": {"summary": report['summary']}}
         if 'details' in report:
             report_dict["storage"]["measures_to_process"] = report['details']
@@ -1402,6 +1465,7 @@ class V1Controller(object):
             "metric": MetricsController(),
             "batch": BatchController(),
             "resource": ResourcesByTypeController(),
+            "resource_type": ResourceTypesController(),
             "aggregation": AggregationController(),
             "capabilities": CapabilityController(),
             "status": StatusController(),
