@@ -20,11 +20,11 @@ import tempfile
 import threading
 import time
 from unittest import case
-import uuid
 import warnings
 
 from gabbi import fixture
-import sqlalchemy.engine.url as sqlalchemy_url
+from oslo_config import cfg
+from oslo_middleware import cors
 import sqlalchemy_utils
 
 from gnocchi import indexer
@@ -82,6 +82,13 @@ class ConfigFixture(fixture.GabbiFixture):
                           os.path.abspath('etc/gnocchi/api-paste.ini'),
                           'api')
 
+        # NOTE(sileht): This is not concurrency safe, but only this tests file
+        # deal with cors, so we are fine. set_override don't work because cors
+        # group doesn't yet exists, and we the CORS middleware is created it
+        # register the option and directly copy value of all configurations
+        # options making impossible to override them properly...
+        cfg.set_defaults(cors.CORS_OPTS, allowed_origin="http://foobar.com")
+
         self.conf = conf
         self.tmp_dir = data_tmp_dir
 
@@ -103,14 +110,11 @@ class ConfigFixture(fixture.GabbiFixture):
 
         # NOTE(jd) All of that is still very SQL centric but we only support
         # SQL for now so let's say it's good enough.
-        url = sqlalchemy_url.make_url(
-            sqlalchemy.SQLAlchemyIndexer.dress_url(
-                conf.indexer.url))
-
-        url.database = url.database + str(uuid.uuid4()).replace('-', '')
-        db_url = str(url)
-        conf.set_override('url', db_url, 'indexer')
-        sqlalchemy_utils.create_database(db_url)
+        conf.set_override(
+            'url',
+            sqlalchemy.SQLAlchemyIndexer._create_new_database(
+                conf.indexer.url),
+            'indexer')
 
         index = indexer.get_driver(conf)
         index.connect()
@@ -170,7 +174,9 @@ class MetricdThread(threading.Thread):
 
     def run(self):
         while self.flag:
-            self.storage.process_background_tasks(self.index)
+            metrics = self.storage.list_metric_with_measures_to_process(
+                None, None, full=True)
+            self.storage.process_background_tasks(self.index, metrics)
             time.sleep(0.1)
 
     def stop(self):
