@@ -17,12 +17,18 @@ import datetime
 import operator
 import uuid
 
+import mock
+
 from gnocchi import archive_policy
 from gnocchi import indexer
 from gnocchi.tests import base as tests_base
 from gnocchi import utils
 
 
+class MockException(Exception):
+    pass
+
+
 class TestIndexer(tests_base.TestCase):
     def test_get_driver(self):
         driver = indexer.get_driver(self.conf)
@@ -50,6 +56,55 @@ class TestIndexerDriver(tests_base.TestCase):
                 {u'granularity': 86400, u'points': 30, u'timespan': 2592000}],
             'name': u'low'}, dict(ap))
 
+    def test_update_archive_policy(self):
+        self.assertRaises(indexer.UnsupportedArchivePolicyChange,
+                          self.index.update_archive_policy, "low",
+                          [archive_policy.ArchivePolicyItem(granularity=300,
+                                                            points=10)])
+        self.assertRaises(indexer.UnsupportedArchivePolicyChange,
+                          self.index.update_archive_policy, "low",
+                          [archive_policy.ArchivePolicyItem(granularity=300,
+                                                            points=12),
+                           archive_policy.ArchivePolicyItem(granularity=3600,
+                                                            points=12),
+                           archive_policy.ArchivePolicyItem(granularity=5,
+                                                            points=6)])
+        apname = str(uuid.uuid4())
+        self.index.create_archive_policy(archive_policy.ArchivePolicy(
+            apname, 0, [(12, 300), (24, 3600), (30, 86400)]))
+        ap = self.index.update_archive_policy(
+            apname, [archive_policy.ArchivePolicyItem(granularity=300,
+                                                      points=6),
+                     archive_policy.ArchivePolicyItem(granularity=3600,
+                                                      points=24),
+                     archive_policy.ArchivePolicyItem(granularity=86400,
+                                                      points=30)])
+        self.assertEqual({
+            'back_window': 0,
+            'aggregation_methods':
+            set(self.conf.archive_policy.default_aggregation_methods),
+            'definition': [
+                {u'granularity': 300, u'points': 6, u'timespan': 1800},
+                {u'granularity': 3600, u'points': 24, u'timespan': 86400},
+                {u'granularity': 86400, u'points': 30, u'timespan': 2592000}],
+            'name': apname}, dict(ap))
+        ap = self.index.update_archive_policy(
+            apname, [archive_policy.ArchivePolicyItem(granularity=300,
+                                                      points=12),
+                     archive_policy.ArchivePolicyItem(granularity=3600,
+                                                      points=24),
+                     archive_policy.ArchivePolicyItem(granularity=86400,
+                                                      points=30)])
+        self.assertEqual({
+            'back_window': 0,
+            'aggregation_methods':
+            set(self.conf.archive_policy.default_aggregation_methods),
+            'definition': [
+                {u'granularity': 300, u'points': 12, u'timespan': 3600},
+                {u'granularity': 3600, u'points': 24, u'timespan': 86400},
+                {u'granularity': 86400, u'points': 30, u'timespan': 2592000}],
+            'name': apname}, dict(ap))
+
     def test_delete_archive_policy(self):
         name = str(uuid.uuid4())
         self.index.create_archive_policy(
@@ -77,10 +132,21 @@ class TestIndexerDriver(tests_base.TestCase):
         self.index.create_archive_policy_rule('rule2', 'abc.xyz.*', name)
         self.index.create_archive_policy_rule('rule3', 'abc.xyz', name)
         rules = self.index.list_archive_policy_rules()
-        self.assertEqual(3, len(rules))
-        self.assertEqual('abc.xyz.*', rules[0]['metric_pattern'])
-        self.assertEqual('abc.xyz', rules[1]['metric_pattern'])
-        self.assertEqual('abc.*', rules[2]['metric_pattern'])
+        # NOTE(jd) The test is not isolated, there might be more than 3 rules
+        found = 0
+        for r in rules:
+            if r['metric_pattern'] == 'abc.xyz.*':
+                found = 1
+            if found == 1 and r['metric_pattern'] == 'abc.xyz':
+                found = 2
+            if found == 2 and r['metric_pattern'] == 'abc.*':
+                break
+        else:
+            self.fail("Metric patterns are not ordered")
+
+        # Ensure we can't delete the archive policy
+        self.assertRaises(indexer.ArchivePolicyInUse,
+                          self.index.delete_archive_policy, name)
 
     def test_create_metric(self):
         r1 = uuid.uuid4()
@@ -91,6 +157,7 @@ class TestIndexerDriver(tests_base.TestCase):
         self.assertEqual(m.created_by_user_id, user)
         self.assertEqual(m.created_by_project_id, project)
         self.assertIsNone(m.name)
+        self.assertIsNone(m.unit)
         self.assertIsNone(m.resource_id)
         m2 = self.index.list_metrics(id=r1)
         self.assertEqual([m], m2)
@@ -170,52 +237,6 @@ class TestIndexerDriver(tests_base.TestCase):
         m = self.index.list_metrics(id=rc.metrics[0].id)
         self.assertEqual(m[0], rc.metrics[0])
 
-    def _do_test_create_instance(self, server_group=None, image_ref=None):
-        r1 = uuid.uuid4()
-        user = str(uuid.uuid4())
-        project = str(uuid.uuid4())
-        kwargs = {'server_group': server_group} if server_group else {}
-
-        rc = self.index.create_resource('instance', r1, user, project,
-                                        flavor_id="1",
-                                        image_ref=image_ref,
-                                        host="foo",
-                                        display_name="lol", **kwargs)
-        self.assertIsNotNone(rc.started_at)
-        self.assertIsNotNone(rc.revision_start)
-        self.assertEqual({"id": r1,
-                          "revision_start": rc.revision_start,
-                          "revision_end": None,
-                          "type": "instance",
-                          "created_by_user_id": user,
-                          "created_by_project_id": project,
-                          "user_id": None,
-                          "project_id": None,
-                          "started_at": rc.started_at,
-                          "ended_at": None,
-                          "display_name": "lol",
-                          "server_group": server_group,
-                          "host": "foo",
-                          "image_ref": image_ref,
-                          "flavor_id": "1",
-                          "original_resource_id": None,
-                          "metrics": {}},
-                         rc.jsonify())
-        rg = self.index.get_resource('generic', r1, with_metrics=True)
-        self.assertEqual(rc.id, rg.id)
-        self.assertEqual(rc.revision_start, rg.revision_start)
-        self.assertEqual(rc.metrics, rg.metrics)
-
-    def test_create_instance(self):
-        self._do_test_create_instance(image_ref='http://foo/bar')
-
-    def test_create_instance_with_server_group(self):
-        self._do_test_create_instance('my_autoscaling_group',
-                                      image_ref='http://foo/bar')
-
-    def test_create_instance_without_image_ref(self):
-        self._do_test_create_instance(image_ref=None)
-
     def test_delete_resource(self):
         r1 = uuid.uuid4()
         self.index.create_resource('generic', r1, str(uuid.uuid4()),
@@ -426,30 +447,40 @@ class TestIndexerDriver(tests_base.TestCase):
         self.assertEqual(e1, r.metrics[0].id)
 
     def test_update_resource_attribute(self):
+        mgr = self.index.get_resource_type_schema()
+        resource_type = str(uuid.uuid4())
+        rtype = mgr.resource_type_from_dict(resource_type, {
+            "col1": {"type": "string", "required": True,
+                     "min_length": 2, "max_length": 15}
+        }, 'creating')
         r1 = uuid.uuid4()
         user = str(uuid.uuid4())
         project = str(uuid.uuid4())
-        rc = self.index.create_resource('instance', r1, user, project,
-                                        flavor_id="1",
-                                        image_ref="http://foo/bar",
-                                        host="foo",
-                                        display_name="lol")
-        rc = self.index.update_resource('instance', r1, host="bar")
-        r = self.index.get_resource('instance', r1, with_metrics=True)
+        # Create
+        self.index.create_resource_type(rtype)
+
+        rc = self.index.create_resource(resource_type, r1, user, project,
+                                        col1="foo")
+        rc = self.index.update_resource(resource_type, r1, col1="foo")
+        r = self.index.get_resource(resource_type, r1, with_metrics=True)
         self.assertEqual(rc, r)
 
     def test_update_resource_no_change(self):
+        mgr = self.index.get_resource_type_schema()
+        resource_type = str(uuid.uuid4())
+        rtype = mgr.resource_type_from_dict(resource_type, {
+            "col1": {"type": "string", "required": True,
+                     "min_length": 2, "max_length": 15}
+        }, 'creating')
+        self.index.create_resource_type(rtype)
         r1 = uuid.uuid4()
         user = str(uuid.uuid4())
         project = str(uuid.uuid4())
-        rc = self.index.create_resource('instance', r1, user, project,
-                                        flavor_id="1",
-                                        image_ref="http://foo/bar",
-                                        host="foo",
-                                        display_name="lol")
-        updated = self.index.update_resource('instance', r1, host="foo",
+        rc = self.index.create_resource(resource_type, r1, user, project,
+                                        col1="foo")
+        updated = self.index.update_resource(resource_type, r1, col1="foo",
                                              create_revision=False)
-        r = self.index.list_resources('instance',
+        r = self.index.list_resources(resource_type,
                                       {"=": {"id": r1}},
                                       history=True)
         self.assertEqual(1, len(r))
@@ -460,28 +491,27 @@ class TestIndexerDriver(tests_base.TestCase):
         r1 = uuid.uuid4()
         user = str(uuid.uuid4())
         project = str(uuid.uuid4())
-        self.index.create_resource('instance', r1, user, project,
-                                   flavor_id="1",
-                                   image_ref="http://foo/bar",
-                                   host="foo",
-                                   display_name="lol")
+        self.index.create_resource('generic', r1, user, project)
         self.assertRaises(
             indexer.ResourceValueError,
             self.index.update_resource,
-            'instance', r1,
+            'generic', r1,
             ended_at=utils.datetime_utc(2010, 1, 1, 1, 1, 1))
 
     def test_update_resource_unknown_attribute(self):
+        mgr = self.index.get_resource_type_schema()
+        resource_type = str(uuid.uuid4())
+        rtype = mgr.resource_type_from_dict(resource_type, {
+            "col1": {"type": "string", "required": False,
+                     "min_length": 1, "max_length": 2},
+        }, 'creating')
+        self.index.create_resource_type(rtype)
         r1 = uuid.uuid4()
-        self.index.create_resource('instance', r1, str(uuid.uuid4()),
-                                   str(uuid.uuid4()),
-                                   flavor_id="1",
-                                   image_ref="http://foo/bar",
-                                   host="foo",
-                                   display_name="lol")
+        self.index.create_resource(resource_type, r1,
+                                   str(uuid.uuid4()), str(uuid.uuid4()))
         self.assertRaises(indexer.ResourceAttributeError,
                           self.index.update_resource,
-                          'instance',
+                          resource_type,
                           r1, foo="bar")
 
     def test_update_non_existent_metric(self):
@@ -526,13 +556,7 @@ class TestIndexerDriver(tests_base.TestCase):
         rc = self.index.create_resource('generic', r1, user, project,
                                         metrics={'foo': e1, 'bar': e2})
         self.index.delete_metric(e1)
-        # It can be called twice
-        try:
-            self.index.delete_metric(e1)
-        except indexer.NoSuchMetric:
-            # It's possible that the metric has been expunged by another
-            # parallel test. No worry.
-            pass
+        self.assertRaises(indexer.NoSuchMetric, self.index.delete_metric, e1)
         r = self.index.get_resource('generic', r1, with_metrics=True)
         self.assertIsNotNone(r.started_at)
         self.assertIsNotNone(r.revision_start)
@@ -549,19 +573,25 @@ class TestIndexerDriver(tests_base.TestCase):
                           "type": "generic",
                           "metrics": {'bar': str(e2)}}, r.jsonify())
 
-    def test_delete_instance(self):
+    def test_delete_resource_custom(self):
+        mgr = self.index.get_resource_type_schema()
+        resource_type = str(uuid.uuid4())
+        self.index.create_resource_type(
+            mgr.resource_type_from_dict(resource_type, {
+                "flavor_id": {"type": "string",
+                              "min_length": 1,
+                              "max_length": 20,
+                              "required": True}
+            }, 'creating'))
         r1 = uuid.uuid4()
-        created = self.index.create_resource('instance', r1,
+        created = self.index.create_resource(resource_type, r1,
                                              str(uuid.uuid4()),
                                              str(uuid.uuid4()),
-                                             flavor_id="123",
-                                             image_ref="foo",
-                                             host="dwq",
-                                             display_name="foobar")
-        got = self.index.get_resource('instance', r1, with_metrics=True)
+                                             flavor_id="foo")
+        got = self.index.get_resource(resource_type, r1, with_metrics=True)
         self.assertEqual(created, got)
         self.index.delete_resource(r1)
-        got = self.index.get_resource('instance', r1)
+        got = self.index.get_resource(resource_type, r1)
         self.assertIsNone(got)
 
     def test_list_resources_by_unknown_field(self):
@@ -607,14 +637,14 @@ class TestIndexerDriver(tests_base.TestCase):
         project = str(uuid.uuid4())
         g = self.index.create_resource('generic', r1, user, project,
                                        user, project)
+        mgr = self.index.get_resource_type_schema()
+        resource_type = str(uuid.uuid4())
+        self.index.create_resource_type(
+            mgr.resource_type_from_dict(resource_type, {}, 'creating'))
         r2 = uuid.uuid4()
-        i = self.index.create_resource('instance', r2,
+        i = self.index.create_resource(resource_type, r2,
                                        user, project,
-                                       user, project,
-                                       flavor_id="123",
-                                       image_ref="foo",
-                                       host="dwq",
-                                       display_name="foobar")
+                                       user, project)
         resources = self.index.list_resources(
             'generic',
             attribute_filter={"=": {"user_id": user}},
@@ -672,13 +702,13 @@ class TestIndexerDriver(tests_base.TestCase):
         r1 = uuid.uuid4()
         g = self.index.create_resource('generic', r1,
                                        str(uuid.uuid4()), str(uuid.uuid4()))
+        mgr = self.index.get_resource_type_schema()
+        resource_type = str(uuid.uuid4())
+        self.index.create_resource_type(
+            mgr.resource_type_from_dict(resource_type, {}, 'creating'))
         r2 = uuid.uuid4()
-        i = self.index.create_resource('instance', r2,
-                                       str(uuid.uuid4()), str(uuid.uuid4()),
-                                       flavor_id="123",
-                                       image_ref="foo",
-                                       host="dwq",
-                                       display_name="foobar")
+        i = self.index.create_resource(resource_type, r2,
+                                       str(uuid.uuid4()), str(uuid.uuid4()))
         resources = self.index.list_resources('generic')
         self.assertGreaterEqual(len(resources), 2)
         g_found = False
@@ -694,7 +724,7 @@ class TestIndexerDriver(tests_base.TestCase):
         else:
             self.fail("Some resources were not found")
 
-        resources = self.index.list_resources('instance')
+        resources = self.index.list_resources(resource_type)
         self.assertGreaterEqual(len(resources), 1)
         for r in resources:
             if r.id == r2:
@@ -713,9 +743,19 @@ class TestIndexerDriver(tests_base.TestCase):
             'generic',
             attribute_filter={"=": {"id": "f00bar" * 50}})
 
-    def test_list_resource_instance_flavor_id_numeric(self):
+    def test_list_resource_attribute_type_numeric(self):
+        """Test that we can pass an integer to filter on a string type."""
+        mgr = self.index.get_resource_type_schema()
+        resource_type = str(uuid.uuid4())
+        self.index.create_resource_type(
+            mgr.resource_type_from_dict(resource_type, {
+                "flavor_id": {"type": "string",
+                              "min_length": 1,
+                              "max_length": 20,
+                              "required": False},
+            }, 'creating'))
         r = self.index.list_resources(
-            'instance', attribute_filter={"=": {"flavor_id": 1.0}})
+            resource_type, attribute_filter={"=": {"flavor_id": 1.0}})
         self.assertEqual(0, len(r))
 
     def test_list_resource_weird_date(self):
@@ -793,7 +833,7 @@ class TestIndexerDriver(tests_base.TestCase):
             key=operator.itemgetter("revision_start"))
         self.assertEqual([r1, r2], resources)
 
-    def test_list_resources_instance_with_history(self):
+    def test_list_resources_custom_with_history(self):
         e1 = uuid.uuid4()
         e2 = uuid.uuid4()
         rid = uuid.uuid4()
@@ -802,6 +842,14 @@ class TestIndexerDriver(tests_base.TestCase):
         new_user = str(uuid.uuid4())
         new_project = str(uuid.uuid4())
 
+        mgr = self.index.get_resource_type_schema()
+        resource_type = str(uuid.uuid4())
+        self.index.create_resource_type(
+            mgr.resource_type_from_dict(resource_type, {
+                "col1": {"type": "string", "required": True,
+                         "min_length": 2, "max_length": 15}
+            }, 'creating'))
+
         self.index.create_metric(e1, user, project,
                                  archive_policy_name="low")
         self.index.create_metric(e2, user, project,
@@ -809,17 +857,14 @@ class TestIndexerDriver(tests_base.TestCase):
         self.index.create_metric(uuid.uuid4(), user, project,
                                  archive_policy_name="low")
 
-        r1 = self.index.create_resource('instance', rid, user, project,
+        r1 = self.index.create_resource(resource_type, rid, user, project,
                                         user, project,
-                                        flavor_id="123",
-                                        image_ref="foo",
-                                        host="dwq",
-                                        display_name="foobar_history",
+                                        col1="foo",
                                         metrics={'foo': e1, 'bar': e2}
                                         ).jsonify()
-        r2 = self.index.update_resource('instance', rid, user_id=new_user,
+        r2 = self.index.update_resource(resource_type, rid, user_id=new_user,
                                         project_id=new_project,
-                                        host="other",
+                                        col1="bar",
                                         append_metrics=True).jsonify()
 
         r1['revision_end'] = r2['revision_start']
@@ -828,8 +873,8 @@ class TestIndexerDriver(tests_base.TestCase):
                           'bar': str(e2)}, r2['metrics'])
         self.assertEqual(new_user, r2['user_id'])
         self.assertEqual(new_project, r2['project_id'])
-        self.assertEqual('other', r2['host'])
-        resources = self.index.list_resources('instance', history=True,
+        self.assertEqual('bar', r2['col1'])
+        resources = self.index.list_resources(resource_type, history=True,
                                               details=False,
                                               attribute_filter={
                                                   "=": {"id": rid}})
@@ -851,12 +896,12 @@ class TestIndexerDriver(tests_base.TestCase):
             started_at=utils.datetime_utc(2000, 1, 1, 23, 23, 23),
             ended_at=utils.datetime_utc(2000, 1, 3, 23, 23, 23))
         r2 = uuid.uuid4()
+        mgr = self.index.get_resource_type_schema()
+        resource_type = str(uuid.uuid4())
+        self.index.create_resource_type(
+            mgr.resource_type_from_dict(resource_type, {}, 'creating'))
         i = self.index.create_resource(
-            'instance', r2, user, project,
-            flavor_id="123",
-            image_ref="foo",
-            host="dwq",
-            display_name="foobar",
+            resource_type, r2, user, project,
             started_at=utils.datetime_utc(2000, 1, 1, 23, 23, 23),
             ended_at=utils.datetime_utc(2000, 1, 4, 23, 23, 23))
         resources = self.index.list_resources(
@@ -882,7 +927,7 @@ class TestIndexerDriver(tests_base.TestCase):
             self.fail("Some resources were not found")
 
         resources = self.index.list_resources(
-            'instance',
+            resource_type,
             attribute_filter={
                 ">=": {
                     "started_at": datetime.datetime(2000, 1, 1, 23, 23, 23)
@@ -977,3 +1022,116 @@ class TestIndexerDriver(tests_base.TestCase):
         self.index.delete_metric(e1)
         metrics = self.index.list_metrics()
         self.assertNotIn(e1, [m.id for m in metrics])
+
+    def test_resource_type_crud(self):
+        mgr = self.index.get_resource_type_schema()
+        rtype = mgr.resource_type_from_dict("indexer_test", {
+            "col1": {"type": "string", "required": True,
+                     "min_length": 2, "max_length": 15}
+        }, "creating")
+
+        # Create
+        self.index.create_resource_type(rtype)
+        self.assertRaises(indexer.ResourceTypeAlreadyExists,
+                          self.index.create_resource_type,
+                          rtype)
+
+        # Get
+        rtype = self.index.get_resource_type("indexer_test")
+        self.assertEqual("indexer_test", rtype.name)
+        self.assertEqual(1, len(rtype.attributes))
+        self.assertEqual("col1", rtype.attributes[0].name)
+        self.assertEqual("string", rtype.attributes[0].typename)
+        self.assertEqual(15, rtype.attributes[0].max_length)
+        self.assertEqual(2, rtype.attributes[0].min_length)
+        self.assertEqual("active", rtype.state)
+
+        # List
+        rtypes = self.index.list_resource_types()
+        for rtype in rtypes:
+            if rtype.name == "indexer_test":
+                break
+        else:
+            self.fail("indexer_test not found")
+
+        # Test resource itself
+        rid = uuid.uuid4()
+        self.index.create_resource("indexer_test", rid,
+                                   str(uuid.uuid4()),
+                                   str(uuid.uuid4()),
+                                   col1="col1_value")
+        r = self.index.get_resource("indexer_test", rid)
+        self.assertEqual("indexer_test", r.type)
+        self.assertEqual("col1_value", r.col1)
+
+        # Deletion
+        self.assertRaises(indexer.ResourceTypeInUse,
+                          self.index.delete_resource_type,
+                          "indexer_test")
+        self.index.delete_resource(rid)
+        self.index.delete_resource_type("indexer_test")
+
+        # Ensure it's deleted
+        self.assertRaises(indexer.NoSuchResourceType,
+                          self.index.get_resource_type,
+                          "indexer_test")
+
+        self.assertRaises(indexer.NoSuchResourceType,
+                          self.index.delete_resource_type,
+                          "indexer_test")
+
+    def _get_rt_state(self, name):
+        return self.index.get_resource_type(name).state
+
+    def test_resource_type_unexpected_creation_error(self):
+        mgr = self.index.get_resource_type_schema()
+        rtype = mgr.resource_type_from_dict("indexer_test_fail", {
+            "col1": {"type": "string", "required": True,
+                     "min_length": 2, "max_length": 15}
+        }, "creating")
+
+        states = {'before': None,
+                  'after': None}
+
+        def map_and_create_mock(rt, conn):
+            states['before'] = self._get_rt_state("indexer_test_fail")
+            raise MockException("boom!")
+
+        with mock.patch.object(self.index._RESOURCE_TYPE_MANAGER,
+                               "map_and_create_tables",
+                               side_effect=map_and_create_mock):
+            self.assertRaises(MockException,
+                              self.index.create_resource_type,
+                              rtype)
+            states['after'] = self._get_rt_state('indexer_test_fail')
+
+        self.assertEqual([('after', 'creation_error'),
+                          ('before', 'creating')],
+                         sorted(states.items()))
+
+    def test_resource_type_unexpected_deleting_error(self):
+        mgr = self.index.get_resource_type_schema()
+        rtype = mgr.resource_type_from_dict("indexer_test_fail2", {
+            "col1": {"type": "string", "required": True,
+                     "min_length": 2, "max_length": 15}
+        }, "creating")
+        self.index.create_resource_type(rtype)
+
+        states = {'before': None,
+                  'after': None}
+
+        def map_and_create_mock(rt, conn):
+            states['before'] = self._get_rt_state("indexer_test_fail2")
+            raise MockException("boom!")
+
+        with mock.patch.object(self.index._RESOURCE_TYPE_MANAGER,
+                               "unmap_and_delete_tables",
+                               side_effect=map_and_create_mock):
+            self.assertRaises(MockException,
+                              self.index.delete_resource_type,
+                              rtype.name)
+            states['after'] = self._get_rt_state('indexer_test_fail2')
+
+        self.assertEqual([('after', 'deletion_error'),
+                          ('before', 'deleting')],
+                         sorted(states.items()))
